# Copyright (c) 2018 UniquID

import click
import json
from uniquid.core.login_manager import LoginManager
from uniquid.core.cli_console import CliConsole
from uniquid.core.shares import Shares
import uniquid.core.constants as constants
import uniquid.commands.common as common


@click.command(name='list-shares')
@click.option('--direction',
              default=constants.SHARE_DIR_BOTH,
              type=click.Choice(constants.SHARE_DIR_ALL),
              help=('Direction of sharing of devices to be listed.'),
              required=False)
@click.option('--org',
              default=None,
              type=click.STRING,
              help=('Organization ID to filter shared devices by.'),
              required=False)
@click.option('--output',
              default=constants.FORMAT_TEXT,
              type=click.Choice(constants.FORMAT_ALL),
              help=('Format used to print data to the console. Valid options'
                    ' are: ') + str(constants.FORMAT_ALL),
              required=False)
def list_shares(direction, org, output):
    """Show a list of all of the shared devices.

    Devices can be those you have shared with other organizations or
    those shared with you by other organizations.
    """
    cc = CliConsole(click.echo, common.print_error,
                    output, click.ClickException)
    lm = LoginManager(cc)
    shares = Shares(cc, lm)
    shares.list_shares(direction, org)


@click.command(name='create-shares')
@click.option('--input-json',
              default=None,
              type=click.STRING,
              help=('JSON object describing the share.'),
              required=False)
@click.option('--input-json-file',
              default=None,
              type=click.File(mode='r'),
              help=('Path to file containing a list of JSON'
                    ' share objects.'),
              required=False)
def create_shares(input_json, input_json_file):
    """Share multiple devices with another Uniquid Organization.

    Devices to be shared must be owned by the user.
    """
    if (input_json is None) == (input_json_file is None):
        raise click.ClickException(constants.ERR_SINGLE_OPTION)
    cc = CliConsole(click.echo, common.print_error,
                    constants.FORMAT_TEXT, click.ClickException)
    lm = LoginManager(cc)
    shares = Shares(cc, lm)
    json_string = input_json
    if input_json_file:
        json_string = input_json_file.read()
    shares.create_shares(json_string)


@click.command(name='delete-shares')
@click.argument('shareids', nargs=-1)  # variadic
@click.option('--input-json',
              default=None,
              type=click.STRING,
              help=('JSON array with IDs of shares to be deleted.'),
              required=False)
@click.option('--input-json-file',
              default=None,
              type=click.File(mode='r'),
              help=('Path to file containing a JSON array of the IDs'
                    ' of the shares to be deleted.'),
              required=False)
def delete_shares(shareids, input_json, input_json_file):
    """Remove the sharing of devices which have Share IDs SHAREIDS.

        Devices to be shared must be owned by the user therefore the devices
        which will have their sharing status removed must be owned by the user.
    """
    if ((input_json is None) == (input_json_file is None) and
            (shareids is None or len(shareids) == 0)):
        raise click.ClickException(constants.ERR_SINGLE_OPTION)
    cc = CliConsole(click.echo, common.print_error,
                    constants.FORMAT_TEXT, click.ClickException)
    lm = LoginManager(cc)
    shares = Shares(cc, lm)
    json_string = input_json
    if input_json_file:
        json_string = input_json_file.read()
    # if user passed one+ arguments, these take priority over a JSON array.
    if(len(shareids) > 0):
        id_list = list()
        for id in shareids:
            id_list.append(str(id))
        json_string = json.dumps(id_list)
    shares.delete_shares(json_string)



# Copyright (c) 2018 UniquID

"""Constant definitions."""

# application version number
APP_VERSION = '0.3.0'

# minimum supported version of python
MIN_VERSION_MAJOR = 3
MIN_VERSION_MINOR = 6

# maximum supported version of credentials file
MAX_VERSION_CREDENTIALS = 1

# permanent Uniquid HTTP API address
# see COR-330: IP address and Port Number of Onboarding HTTP API.
UNIQUID_LOGIN_IP = '52.47.97.110'
UNIQUID_LOGIN_PORT = '3000'
UNIQUID_URL_SCHEME = 'http://'
ORCHESTRATOR_PORT = '8080'

# error messages for authentication and login
ERR_LOGGED_OUT = 'User is logged out. Please login again.'
ERR_LOGIN_REJECTED = 'Login rejected. Please check user credentials.'
ERR_LOGOUT_REJECTED = 'Logout rejected. Session may have expired.'
ERR_LOGOUT_ERROR = 'Failed to logout. May already be logged out.'
ERR_ORG_NOT_FOUND = 'Organization not registered at Uniquid: '
ERR_QUERY_ERROR = 'Server returned bad data.'
ERR_MISSING_API_KEY = 'Login fail. Missing API key.'
ERR_MISSING_USERNAME = 'Login fail. Missing Username or email.'
ERR_MISSING_ORG_ID = 'Login fail. Missing Organization ID.'
ERR_MISSING_ORG_PORT = 'Login fail. Missing Organization IP/port.'
ERR_CRED_ORG = 'Login fail. Cannot specify credential file and organization.'
ERR_CRED_USER = 'Login fail. Cannot specify credential file and user.'
ERR_CRED_KEY = 'Login fail. Cannot specify credential file and access key.'
ERR_ORG_ID_URL = 'Login fail. Cannot specify organization identifier and URL.'
ERR_LOGIN_ORG_URL = 'Login fail. Specified login URL and organization URL.'
ERR_BAD_ORCH_URL = 'Login fail. Incorrect API URL.'
ERR_MISSING_IP_PORT = 'Login fail. Require both HTTP API IP and port number.'
ERR_NO_RESPONSE = 'Connection fail. No response from server.'
ERR_INVALID_URL = 'Login fail. Invalid Uniquid URL as argument.'
ERR_ORCHESTRATOR_SPINUP = 'Login suspended.'
ERR_FILE_PERM = "Insufficient permissions to open file: "
ERR_FILE_OPEN = "Login fail. Missing credentials file: "
# error messages for commands
ERR_ENDPOINT_REJECTED = 'Server rejected credentials. Please login again.'
ERR_SERVER_ERROR = 'Server internal error. Please try again.'
ERR_SINGLE_OPTION = 'One option/argument must be passed.'
ERR_INVALID_JSON = 'Invalid format of JSON data.'
ERR_MISSING_TXID = 'TXID must be passed to command.'
ERR_NO_CONTRACT = 'Contract not found.'
ERR_NO_SHARE = 'Share not found. Share Id: '
# error messages for file system
ERR_TEMP_FILE = ('Failed to create or access a temporary file. ' +
                 'Check directory permissions.')
ERR_BAD_CRED_FILE = 'Invalid credentials file.'
ERR_CRED_FILE_VERSION = 'Unsupported version of credentials file.'
# unknown error - shouldn't happen
ERR_UNKNOWN_ERROR = 'Error occurred. Please try again.'

# supported formats for data output to the console
FORMAT_TEXT = 'text'
FORMAT_JSON = 'json'
FORMAT_ALL = [FORMAT_TEXT, FORMAT_JSON]

# User dialogs
TXT_PREFIX_OK = 'Success. '
TXT_PREFIX_FAIL = 'Failure. '
TXT_LOGIN_OK = TXT_PREFIX_OK + 'Logged in: '
TXT_LOGOUT_OK = TXT_PREFIX_OK + 'Logged out: '
TXT_CONTRACT_CREATED = TXT_PREFIX_OK + 'Contract(s) requested: '
TXT_CONTRACT_CREATED_ALL = TXT_PREFIX_OK + 'All contracts requested.'
TXT_CONTRACT_INTERRUPTED = (TXT_PREFIX_FAIL +
                            'Requests interrupted. Contract(s) requested: ')
TXT_CONTRACT_DELETED = TXT_PREFIX_OK + 'Contract deleted. TXID: '
TXT_CONTRACT_DELETED_ALL = TXT_PREFIX_OK + 'All contracts deleted.'
TXT_SHARE_CREATED = TXT_PREFIX_OK + 'Devices shared: '
TXT_SHARE_INTERRUPTED = (TXT_PREFIX_FAIL +
                         'Share creation interrupted. Devices shared: ')
TXT_SHARE_DEL_INTERRUPTED = (TXT_PREFIX_FAIL +
                             ('Share deletion interrupted. '
                              ' Device share(s) deleted: '))
TXT_SHARE_DELETED = TXT_PREFIX_OK + ' Share(s) deleted: '
TXT_LOGIN_FAIL_ORCH_SPINUP = (TXT_PREFIX_FAIL +
                              ('System is starting up. Please try again after '
                               'a brief wait.'))
TXT_CONN_STATUS = 'Connection status: '
TXT_CONN_LOGGED_IN = 'logged in'
TXT_CONN_LOGGED_OUT = 'logged out'

# device sharing directions
SHARE_DIR_IN = 'in'
SHARE_DIR_OUT = 'out'
SHARE_DIR_BOTH = 'all'
SHARE_DIR_ALL = [SHARE_DIR_IN, SHARE_DIR_OUT, SHARE_DIR_BOTH]

# keys of JSON objects used in Orchestrator HTTP API
KEY_ORG_ID = 'orgId'

# credentials file schema
CREDENTIALS_SCHEMA = {
    'type': 'object',
    'properties': {
        'version': {'type': 'number'},
        'organization': {'type': 'string'},
        'user': {'type': 'string'},
        'accessKey': {'type': 'string'}
    },
    'required': ['version', 'organization', 'user', 'accessKey']
}



# Copyright (c) 2018 UniquID

"""JSON Schema definitions."""

NEW_CONTRACT_SCHEMA = {
    'type': 'array',
    'items': {
        'type': 'object',
        'properties': {
            'provider': {'type': 'string'},
            'user': {'type': 'string'},
            'functions': {'type': 'array',
                          'items': {'type': 'number'}}
        },
        'required': ['provider', 'user']
    },
    'minItems': 1
}

DELETE_CONTRACT_SCHEMA = {
    'type': 'array',
    'items': {'type': 'string'},
    'minItems': 1
}

NEW_SHARE_SCHEMA = {
    'type': 'array',
    'items': {
        'type': 'object',
        'properties': {
            'orgId': {'type': 'string'},
            'xpub': {'type': 'string'}
        },
        'required': ['orgId', 'xpub']
    },
    'minItems': 1
}

DELETE_SHARES_SCHEMA = {
    'type': 'array',
    'items': {'type': 'string'},
    'minItems': 1
}



# Copyright (c) 2018 UniquID

import requests
import uniquid.core.constants as constants
import uniquid.core.helper as helper


class Devices:
    """Implements all commands related to devices.
    """

    def __init__(self, in_console, in_login_mgr):
        """Initialise an instance of the class."""
        self._console = in_console
        self._login_mgr = in_login_mgr

    def show_device(self, in_xpub):
        """Show detailed information for one device.

            Arguments:
            in_xpub -- Public key which uniquely identifies a device.
        """
        # check that user is logged in
        session_key = self._login_mgr.get_session_key()
        if session_key is None:
            self._console.exception(constants.ERR_LOGGED_OUT)
        # retrieve the device info from the server using HTTP API
        url_string = 'http://{0}:{1}/api/v1/devices/{2}'
        url = url_string.format(self._login_mgr.get_api_ip(),
                                self._login_mgr.get_api_port(),
                                in_xpub)
        cookies = dict(JSESSIONID=session_key)
        try:
            response = requests.get(url, cookies=cookies)
            if (response.status_code == 200
                    and response.json()):
                self._console.print_objects(response.json())
            elif response.status_code == 404:
                # not an error to query a device which doesn't exist
                self._console.ok('')
            elif response.status_code == 401:
                self._console.exception(constants.ERR_LOGIN_REJECTED)
            elif response.status_code == 500:
                self._console.exception(constants.ERR_SERVER_ERROR)
            else:
                self._console.exception(constants.ERR_UNKNOWN_ERROR
                                        + ' ['+str(response.status_code) + ']')
        except requests.exceptions.ConnectionError:
            self._console.exception(constants.ERR_NO_RESPONSE)

    def list_devices(self, in_sort_by, in_is_sort_asc):
        """List all devices.

        Arguments:
        in_sort_by -- Key of field to sort list by.
        in_is_sort_asc -- Boolean. True if sort is ascending, otherwise False.
        """
        # check that user is logged in
        session_key = self._login_mgr.get_session_key()
        if session_key is None:
            self._console.exception(constants.ERR_LOGGED_OUT)
        # retrieve the device info from the server using HTTP API
        url_string = 'http://{0}:{1}/api/v1/devices'
        url = url_string.format(self._login_mgr.get_api_ip(),
                                self._login_mgr.get_api_port())
        cookies = dict(JSESSIONID=session_key)
        try:
            response = requests.get(url, cookies=cookies)
            if (response.status_code == 200
                    and isinstance(response.json(), list)):
                sorted = list()
                if in_sort_by:
                    sorted.extend(helper.sort_list(response.json(),
                                                   in_sort_by,
                                                   in_is_sort_asc,
                                                   True))
                else:
                    sorted.extend(response.json())
                self._console.print_objects(sorted)
            elif response.status_code == 401:
                self._console.exception(constants.ERR_LOGIN_REJECTED)
            elif response.status_code == 500:
                self._console.exception(constants.ERR_SERVER_ERROR)
            else:
                self._console.exception(constants.ERR_UNKNOWN_ERROR + ' ['
                                        + str(response.status_code) + ']')
        except requests.exceptions.ConnectionError:
            self._console.exception(constants.ERR_NO_RESPONSE)



# Copyright (c) 2018 UniquID

import click


def print_error(in_string):
    """Print a string to stderr using click framework. Provides a platform
    independant way of printing to stderr to the core code.

        Arguments:
        in_string -- String.
    """
    click.echo(message=in_string, err=True)



# Copyright (c) 2018 UniquID

import click
from uniquid.core.login_manager import LoginManager
from uniquid.core.cli_console import CliConsole
from uniquid.core.devices import Devices
import uniquid.core.constants as constants
import uniquid.commands.common as common


@click.command(name='show-device')
@click.argument('xpub',
                required=True)
@click.option('--output',
              default=constants.FORMAT_TEXT,
              type=click.Choice(constants.FORMAT_ALL),
              help=('Format used to print data to the console. Valid options'
                    ' are: ') + str(constants.FORMAT_ALL),
              required=False)
def show_device(xpub, output):
    """Show information about a single device, identified by XPUB."""
    cc = CliConsole(click.echo, common.print_error,
                    output, click.ClickException)
    lm = LoginManager(cc)
    devices = Devices(cc, lm)
    devices.show_device(xpub)


@click.command(name='list-devices')
@click.option('--sort',
              default=None,
              type=click.STRING,
              help='Field name to sort output in ascending order.',
              required=False)
@click.option('--sort-desc',
              default=None,
              type=click.STRING,
              help='Field name to sort output in descending order.',
              required=False)
@click.option('--output',
              default=constants.FORMAT_TEXT,
              type=click.Choice(constants.FORMAT_ALL),
              help=('Format used to print data to the console. Valid options'
                    ' are: ') + str(constants.FORMAT_ALL),
              required=False)
def list_devices(sort, sort_desc, output):
    """Show a list of all devices."""
    cc = CliConsole(click.echo, common.print_error,
                    output, click.ClickException)
    lm = LoginManager(cc)
    devices = Devices(cc, lm)
    sort_by = None
    is_sort_asc = True
    if sort is not None:
        sort_by = sort
    elif sort_desc:
        sort_by = sort_desc
        is_sort_asc = False
    devices.list_devices(sort_by, is_sort_asc)



# Copyright (c) 2018 UniquID

import json
import requests
import jsonschema
import uniquid.core.constants as constants
import uniquid.core.schema as schema
import uniquid.core.helper as helper


class Shares:
    """Implements all commands related to device sharing across orgs.
    """

    def __init__(self, in_console, in_login_mgr):
        """Initialise an instance of the class."""
        self._console = in_console
        self._login_mgr = in_login_mgr

    def list_shares(self, in_direction, in_org):
        """List all shared devices.

        Function is complex because of the way the :orgId routing
        parameter can be used in the /api/v1/shares endpoint. Where possible,
        function trys to use the Orchestrator to filter based on orgId
        rather than filtering in the client.

        Arguments:
        in_direction -- Direction of sharing of the devices.
        in_org -- ID of a Organization to filter devices by.
        """
        # check that user is logged in
        session_key = self._login_mgr.get_session_key()
        if session_key is None:
            self._console.exception(constants.ERR_LOGGED_OUT)
        # retrieve the device info from the server using HTTP API. if
        # possible, filter by org Id on server.
        org_string = ''
        if (in_direction == constants.SHARE_DIR_IN and
                in_org and
                len(in_org) > 0):
            org_string = in_org
        elif (in_direction == constants.SHARE_DIR_OUT):
            org_string = self._login_mgr.get_org_id()
        elif (in_direction == constants.SHARE_DIR_BOTH):
            pass
        # retrieve share objects from server
        try:
            url_string = 'http://{0}:{1}/api/v1/shares/{2}'
            url = url_string.format(self._login_mgr.get_api_ip(),
                                    self._login_mgr.get_api_port(),
                                    org_string)
            cookies = dict(JSESSIONID=session_key)
            response = requests.get(url, cookies=cookies)
            if (response.status_code == 200
                    and isinstance(response.json(), list)):
                # filter the items based on sharing direction
                filtered = list()
                if in_direction != constants.SHARE_DIR_BOTH:
                        filtered = helper.filter_list(response.json(),
                                                      'shareType',
                                                      in_direction)
                else:
                    filtered = response.json()
                # when listing devices which are shared both inwards and
                # outwards and filtering on org id, must filter both the
                # orgId and the ownerId.
                if (in_direction == constants.SHARE_DIR_BOTH and
                        in_org and
                        len(in_org) > 0):
                    filtered_in = helper.filter_list(filtered,
                                                     'ownerId',
                                                     in_org)
                    filtered_out = helper.filter_list(filtered,
                                                      constants.KEY_ORG_ID,
                                                      in_org)
                    filtered.clear()
                    filtered.extend(filtered_in)
                    filtered.extend(filtered_out)
                elif (in_direction == constants.SHARE_DIR_OUT and
                      in_org and
                      len(in_org) > 0):
                    filtered_out = helper.filter_list(filtered,
                                                      constants.KEY_ORG_ID,
                                                      in_org)
                    filtered.clear()
                    filtered.extend(filtered_out)
                self._console.print_objects(filtered)
            elif response.status_code == 401:
                self._console.exception(constants.ERR_LOGIN_REJECTED)
            elif response.status_code == 500:
                self._console.exception(constants.ERR_SERVER_ERROR)
            else:
                self._console.exception(constants.ERR_UNKNOWN_ERROR + ' ['
                                        + str(response.status_code) + ']')
        except requests.exceptions.ConnectionError:
            self._console.exception(constants.ERR_NO_RESPONSE)

    def create_shares(self, in_json):
        """Request that multiple devices are shared with another org.

            Arguments:
            in_json -- JSON sting describing a list of share object(s).
        """
        if in_json is None:
            self._console.exception(constants.exception('Internal. ' +
                                                        'Missing JSON.'))
        # check that user is logged in
        session_key = self._login_mgr.get_session_key()
        if session_key is None:
            self._console.exception(constants.ERR_LOGGED_OUT)
        # validate json before bothering the backend
        json_data = None
        try:
            json_data = json.loads(in_json)
            jsonschema.validate(json_data, schema.NEW_SHARE_SCHEMA)
            assert isinstance(json_data, list)
        except (jsonschema.exceptions.ValidationError,
                json.decoder.JSONDecodeError) as e:
            self._console.exception(constants.ERR_INVALID_JSON + ': ' + str(e))
        # post json to server
        url_string = 'http://{0}:{1}/api/v1/shares'
        url = url_string.format(self._login_mgr.get_api_ip(),
                                self._login_mgr.get_api_port())
        cookies = dict(JSESSIONID=session_key)
        count = 0
        try:
            for obj in json_data:
                response = requests.post(url, cookies=cookies, json=obj)
                if response.status_code == 200:
                    count = count + 1
                elif response.status_code == 401:
                    self._console.error(constants.TXT_SHARE_INTERRUPTED
                                        + str(count))
                    self._console.exception(constants.ERR_LOGIN_REJECTED)
                elif response.status_code == 500:
                    self._console.error(constants.TXT_SHARE_INTERRUPTED
                                        + str(count))
                    self._console.exception(constants.ERR_SERVER_ERROR)
                else:
                    self._console.error(constants.TXT_SHARE_INTERRUPTED
                                        + str(count))
                    self._console.exception(constants.ERR_UNKNOWN_ERROR + ' ['
                                            + str(response.status_code) + ']')
            self._console.ok(constants.TXT_SHARE_CREATED + str(count))
        except requests.exceptions.ConnectionError:
            self._console.error(constants.TXT_SHARE_INTERRUPTED
                                + str(count))
            self._console.exception(constants.ERR_NO_RESPONSE)

    def delete_shares(self, in_json):
        """Request the removal of sharing of own devices with another org.

        Arguments:
        in_json -- JSON sting describing a list of share object(s).
        """
        if in_json is None:
            self._console.exception(constants.exception('Internal. ' +
                                                        'Missing JSON.'))
        # check that user is logged in
        session_key = self._login_mgr.get_session_key()
        if session_key is None:
            self._console.exception(constants.ERR_LOGGED_OUT)
        # validate json before bothering the backend
        json_data = None
        try:
            json_data = json.loads(in_json)
            jsonschema.validate(json_data, schema.DELETE_SHARES_SCHEMA)
            assert isinstance(json_data, list)
        except (jsonschema.exceptions.ValidationError,
                json.decoder.JSONDecodeError) as e:
            self._console.exception(constants.ERR_INVALID_JSON + ': ' + str(e))
        # post json to server
        url_string = 'http://{0}:{1}/api/v1/shares/{2}'
        cookies = dict(JSESSIONID=session_key)
        count = 0
        try:
            for share_id in json_data:
                url = url_string.format(self._login_mgr.get_api_ip(),
                                        self._login_mgr.get_api_port(),
                                        str(share_id))
                response = requests.delete(url, cookies=cookies)
                if response.status_code == 200:
                    count = count + 1
                elif response.status_code == 401:
                    self._console.error(constants.TXT_SHARE_DEL_INTERRUPTED
                                        + str(count))
                    self._console.exception(constants.ERR_LOGIN_REJECTED)
                elif response.status_code == 404:
                    self._console.error(constants.TXT_SHARE_DEL_INTERRUPTED
                                        + str(count))
                    self._console.exception(constants.ERR_NO_SHARE + share_id)
                elif response.status_code == 500:
                    self._console.error(constants.TXT_SHARE_DEL_INTERRUPTED
                                        + str(count))
                    self._console.exception(constants.ERR_SERVER_ERROR)
                else:
                    self._console.error(constants.TXT_SHARE_DEL_INTERRUPTED
                                        + str(count))
                    self._console.exception(constants.ERR_UNKNOWN_ERROR + ' ['
                                            + str(response.status_code) + ']')
            self._console.ok(constants.TXT_SHARE_DELETED + str(count))
        except requests.exceptions.ConnectionError:
            self._console.error(constants.TXT_SHARE_DEL_INTERRUPTED
                                + str(count))
            self._console.exception(constants.ERR_NO_RESPONSE)



# Copyright (c) 2018 UniquID

import json
import jsonschema
import requests
import tempfile
import os

import uniquid.core.constants as constants
import uniquid.core.helper as helper

# components of name and path of temporary session file
_TEMP_DIR = 'com.uniquid'
_TEMP_FILE_PREFIX = 'login_tempfile_v1'
_TEMP_FILE_SUFFIX = '.tmp'
# path of default credentials file
_DEFAULT_CRED_FILE = os.path.join('~', '.uniquid', 'credentials.json')


class LoginManager:
    """Controls the state of the login process and caches the session key
    between invocations.
    """

    def __init__(self, in_console):
        """Initialise the login manager instance."""
        self._console = in_console
        self._api_ip = None
        self._api_port = None
        self._session_key = None
        self._org_id = None

    def connect(self,
                in_org_id=None,
                in_access_key=None,
                in_user=None,
                in_login_url=None,
                in_org_url=None,
                in_cred_file=None):
        """Connect to the UniquID servers.

            Create a connection with the UniquID servers.

            Retrieve the temporary session key which must be used for
            HTTP calls to the Orchestrator API.

            Arguments:
            in_org_id -- Unique identifier of the Organization.
            in_access_key -- API Key to authenticate with Orchestrator
            HTTP API.
            in_user -- Username in the Orchestrator.
            in_login_url -- IP address and port number of the UniquID
            Onboarding HTTP API. Must be in the
            format <ip address>:<port number>.
            in_org_url -- IP address  and port number of the Orchestrator
            HTTP API. Must be in the format <ip address>:<port number>.
            in_cred_file -- Path to credentials file.
        """
        self._check_args(in_org_id, in_access_key, in_user, in_login_url,
                         in_org_url, in_cred_file)
        # access key and user may come from credentials file
        actual_access_key = in_access_key
        actual_user = in_user
        # check if a credential file exists
        cred_file = None
        credentials = None
        if in_cred_file:
            cred_file = self._get_user_cred_file(in_cred_file)
        else:
            cred_file = self._get_default_cred_file()
        if cred_file:
            credentials = self._read_cred_file(cred_file)
        # must specify user/key or credential file
        if (in_access_key is None and
                credentials is None):
            self._console.exception(constants.ERR_MISSING_API_KEY)
        elif (in_user is None and
                credentials is None):
            self._console.exception(constants.ERR_MISSING_USERNAME)
        # command line organization take precedence
        if in_org_id is not None:
            # user specified an org so start login from beginning
            self._api_ip = None
            self._api_port = None
            self._session_key = None
            self._org_id = in_org_id
            self._write_temp_file()
        else:
            # re-use last org login
            self._read_temp_file()
        # decide which organization to use
        if (in_org_id is None and
                credentials is not None):
            self._org_id = credentials['organization']
            self._write_temp_file()
        # decide which user to use
        if in_user is not None:
            actual_user = in_user
        else:
            actual_user = credentials['user']
        # decide which access key to use
        if in_access_key is not None:
            actual_access_key = in_access_key
        else:
            actual_access_key = credentials['accessKey']
        # address of HTTP API used to query the Orchestrator address
        login_address = None
        # address of orchestrator HTTP API for the organization
        orch_address = None
        try:
            # parse the IP address of the Onboarding HTTP API. Use the value
            # passed as an argument or use a default URL.
            login_address = helper.parse_url(in_login_url,
                                             constants.UNIQUID_LOGIN_IP,
                                             constants.UNIQUID_LOGIN_PORT)
            # retrieve IP address from Onboarding API. if user passes the
            # IP address and port of the Orchestrator, bypass the query to
            # the Onboarding API.
            if in_org_url is None:
                orch_address = self._get_orchestrator(
                                        in_org=self._org_id,
                                        in_login_ip=login_address['ip'],
                                        in_login_port=login_address['port'])
            if orch_address:
                self._api_ip = orch_address.get('publicIp')
                self._api_port = orch_address.get('port',
                                                  constants.ORCHESTRATOR_PORT)
                # check orchestrator is not being started
                if ('status' in orch_address and
                        orch_address['status'] != 'Ready'):
                    self._console.warning(constants.TXT_LOGIN_FAIL_ORCH_SPINUP)
                    self._write_temp_file()
                    self._console.exception(constants.ERR_ORCHESTRATOR_SPINUP)
            elif in_org_url:
                orch_address = helper.parse_url(in_org_url, '', '')
                self._api_ip = orch_address['ip']
                self._api_port = orch_address['port']
            else:
                self._console.exception(constants.ERR_QUERY_ERROR)
        except RuntimeError as e:
            self._console.exception(str(e.args))
        # have a valid orchestrator so try and login
        self._login_orchestrator(actual_user, actual_access_key)
        # store the session state in persistant storage i.e. temp file
        if (self._org_id is not None):
            self._write_temp_file()
            self._console.ok(constants.TXT_LOGIN_OK + self._org_id)
        else:
            self._console.exception(constants.ERR_LOGIN_REJECTED)

    def disconnect(self):
        """Disconnect from the UniquID server and delete cached values."""
        self._read_temp_file()
        if(self._api_ip is None or
                self._api_port is None):
            self._console.exception(constants.ERR_LOGOUT_ERROR)
        # inform server we are diconnecting
        url = 'http://{0}:{1}/api/v1/logout'.format(self._api_ip,
                                                    self._api_port)
        cookies = dict(JSESSIONID=self._session_key)
        response = None
        try:
            response = requests.get(url, cookies=cookies)
        except requests.exceptions.ConnectionError:
            self._console.exception(constants.ERR_NO_RESPONSE)
        if response.status_code == 200:
            self._api_ip = None
            self._api_port = None
            self._session_key = None
            # the Organization ID is kept for future use and not cleared.
            self._write_temp_file()
            self._console.ok(constants.TXT_LOGOUT_OK + str(self._org_id))
        elif response.status_code == 401:
            self._console.exception(constants.ERR_LOGOUT_REJECTED)
        elif response.status_code == 500:
            self._console.exception(constants.ERR_SERVER_ERROR)
        else:
            self._console.exception(constants.ERR_LOGOUT_REJECTED)

    def print_info(self):
        """Print connection and Orchestrator information."""
        self._read_temp_file()
        if(self._api_ip is None or
                self._api_port is None):
            self._console.exception(constants.ERR_LOGOUT_ERROR)
        # request info from server
        url = 'http://{0}:{1}/api/v1/info'.format(self._api_ip,
                                                  self._api_port)
        cookies = dict(JSESSIONID=self._session_key)
        response = None
        try:
            response = requests.get(url, cookies=cookies)
        except requests.exceptions.ConnectionError:
            self._console.exception(constants.ERR_NO_RESPONSE)
        if (response.status_code == 200 and
                response.json()):
            self._console.print_objects(response.json())
            # assemble orchestrator info
            orch_info = dict()
            orch_info['orgIp'] = self.get_api_ip()
            orch_info['orgPort'] = self.get_api_port()
            self._console.print_objects(orch_info)
        elif response.status_code == 401:
            self._console.exception(constants.ERR_LOGIN_REJECTED)
        elif response.status_code == 500:
            self._console.exception(constants.ERR_SERVER_ERROR)
        else:
            self._console.exception(constants.ERR_UNKNOWN_ERROR)

    def get_conn_state(self):
        """Get the state of the connection to the server.

            Returns:
            String describing the connection state.
        """
        retval = constants.TXT_CONN_LOGGED_OUT
        self._read_temp_file()
        if (self._session_key is not None
                and self._api_ip is not None
                and self._api_port is not None):
            retval = constants.TXT_CONN_LOGGED_IN
        return retval

    def get_session_key(self):
        """Get the Orchestrator session key.

            Returns:
            String holding the current saved session key. None is returned
            if there is no valid session key.
        """
        if self._session_key is None:
                self._read_temp_file()
        return self._session_key

    def get_api_ip(self):
        """Get the IP address of the Orchestrator API.

            Returns:
            String with the IP address.
        """
        if self._api_ip is None:
            self._read_temp_file()
        return self._api_ip

    def get_api_port(self):
        """Get the port provided by the Orchestrator.

            Returns:
            String with the port number.
        """
        if self._api_port is None:
            self._read_temp_file()
        return self._api_port

    def get_org_id(self):
        """Get the ID of the user's Organization account.

            Returns:
            String with the Organization ID.
        """
        if self._org_id is None:
            self._read_temp_file()
        return self._org_id

    def _check_args(self,
                    in_org_id=None,
                    in_access_key=None,
                    in_user=None,
                    in_login_url=None,
                    in_org_url=None,
                    in_cred_file=None):
        """Check that the correct combination of arguments and options
            are passed to the function connect. Throws an exception if an
            incorrect argument or option combination is found.
            Arguments:
            Must be the exact same as the connect method.
            Throws:
            Exception if invalid argument combination is specified.
        """
        if (in_cred_file is not None and
                in_org_id is not None):
            self._console.exception(constants.ERR_CRED_ORG)
        elif (in_cred_file is not None and
              in_user is not None):
            self._console.exception(constants.ERR_CRED_USER)
        elif (in_cred_file is not None and
              in_access_key is not None):
            self._console.exception(constants.ERR_CRED_KEY)
        elif (in_org_id is not None and
              in_org_url is not None):
            self._console.exception(constants.ERR_ORG_ID_URL)
        elif (in_login_url is not None and
              in_org_url is not None):
            self._console.exception(constants.ERR_LOGIN_ORG_URL)

    def _get_orchestrator(self,
                          in_org=None,
                          in_login_ip=constants.UNIQUID_LOGIN_IP,
                          in_login_port=constants.UNIQUID_LOGIN_PORT):
        """Get the IP Address and the Port Number of the Orchestrator
            Public HTTP API.

            Arguments:
            in_org -- Organization ID.
            in_login_ip -- IP address of the Onboarding HTTP API. Uses a
            hardcoded IP Address if none is provided by the caller.
            in_login_port -- Port number of the Onboarding HTTP API. Uses a
            hardcoded Port Number if none is provided by the caller.
            Returns:
            JSON object describing IP address of the Orchestrator.
        """
        retval = None
        # must know Organization ID to get Orchestrator IP
        if in_org is None:
            self._console.exception(constants.ERR_MISSING_ORG_ID)
        elif in_login_ip is None:
            self._console.exception('Internal. Missing IP address.')
        elif in_login_port is None:
            self._console.exception('Internal. Missing port number.')
        headers = {'Content-Type': 'application/json'}
        # retrieve the orchestrator address from the server
        url_string = 'http://{0}:{1}/api/v1/instances/{2}'
        url = url_string.format(in_login_ip,
                                in_login_port,
                                in_org)
        try:
            response = requests.get(url, headers=headers)
            if response.status_code == 200:
                try:
                    data = response.json()
                    if (constants.KEY_ORG_ID in data and
                            not data[constants.KEY_ORG_ID] is None):
                        retval = data
                except ValueError:
                    self._console.exception(constants.ERR_QUERY_ERROR)
            elif response.status_code == 401:
                self._console.exception(constants.ERR_LOGIN_REJECTED)
            else:
                self._console.exception(constants.ERR_ORG_NOT_FOUND
                                        + str(self._org_id))
        except requests.exceptions.ConnectionError:
            self._console.exception(constants.ERR_NO_RESPONSE)
        return retval

    def _login_orchestrator(self, in_user, in_access_key):
        """Login to the orchestrator instance for the organization.
        """
        is_orch_resp_ok = False
        # login and retrieve session id
        headers = {'Content-Type': 'application/json'}
        # retrieve the session key from the server
        url = 'http://{0}:{1}/api/v1/login'.format(self._api_ip,
                                                   self._api_port)
        data = {'userId': in_user, 'password': in_access_key}
        response = None
        try:
            response = requests.post(url,
                                     data=json.dumps(data),
                                     headers=headers)
        except requests.exceptions.ConnectionError:
            self._console.exception(constants.ERR_NO_RESPONSE)
        if response.status_code == 200:
            is_orch_resp_ok = True
        elif response.status_code == 401:
            self._console.exception(constants.ERR_LOGIN_REJECTED)
        elif response.status_code == 500:
            self._console.exception(constants.ERR_SERVER_ERROR)
        else:
            self._console.exception(constants.ERR_LOGIN_REJECTED)
        # server must return a valid session key
        if (is_orch_resp_ok
                and response is not None
                and response.headers is not None
                and 'Set-Cookie'in response.headers
                and response.headers['Set-Cookie'] is not None
                and response.cookies is not None):
                    self._session_key = response.cookies['JSESSIONID']
        else:
            self._console.exception(constants.ERR_LOGIN_REJECTED)

    def _read_temp_file(self):
        """Read the saved state from the temporary file.

        Function reads the saved state from the temporary file and
        writes values to data members of this object.
        """
        indata = None
        fd = self._get_temp_file('r')
        if fd is not None:
            with fd as infile:
                indata = json.load(infile)
                if 'session_key' in indata and indata['session_key']:
                    self._session_key = indata['session_key']
                if 'api_ip' in indata and indata['api_ip']:
                    self._api_ip = indata['api_ip']
                if 'api_port' in indata and indata['api_port']:
                    self._api_port = indata['api_port']
                # don't overwrite org id if already in memory
                if (self._org_id is None and
                        'org' in indata and
                        indata['org']):
                    self._org_id = indata['org']
        else:
            # failed to open the temp file
            self._session_key = None
            self._api_ip = None
            self._api_key = None
            self._org_id = None

    def _write_temp_file(self):
        """Write the state of login to temporary file.
        """
        fd = self._get_temp_file('w')
        data = {}
        if self._session_key is not None:
            data['session_key'] = self._session_key
        if self._api_ip is not None:
            data['api_ip'] = self._api_ip
        if self._api_port is not None:
            data['api_port'] = self._api_port
        if self._org_id is not None:
            data['org'] = self._org_id
        with fd as outfile:
            json.dump(data, outfile)
            self._close_temp_file(outfile)

    def _get_temp_path(self):
        """Get the full path of the temporary file used to persist data.

            Returns:
            String with the full path to the temporary file."""
        retval = os.path.join(tempfile.gettempdir(),
                              _TEMP_DIR,
                              _TEMP_FILE_PREFIX + _TEMP_FILE_SUFFIX)
        return retval

    def _get_temp_file(self, in_mode='r'):
        """Get a handle to the temporary file that stores the session key.

            Arguments:
            in_mode -- 'r' to open in read mode. 'w' to open in write mode.
            Returns:
            A valid file handle or None.
        """
        retval = None
        if not (in_mode is 'r' or in_mode is 'w'):
            self._console.exception('Internal: Incorrect file mode specified.')
        # create temp dir if doesn't exist
        dir_path = os.path.dirname(self._get_temp_path())
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)
        # if temp file already exists, use it.
        temp_path = self._get_temp_path()
        try:
            retval = open(temp_path, in_mode)
        except OSError:
            retval = None
        except (IOError, OSError):
            self._console.exception(constants.ERR_TEMP_FILE)
        return retval

    def _close_temp_file(self, in_fd):
        """Close the file handle for the temporary file.

            Arguments:
            in_fd -- File descriptor.
        """
        in_fd.close()

    def _get_default_cred_file(self):
        """Get a handle to the credential file, if one exists, in the
           configuration directory of the user's home directory.
           Returns:
           File handle or None if the file does not exist.
        """
        retval = None
        default_path = os.path.expanduser(_DEFAULT_CRED_FILE)
        try:
            retval = open(default_path, 'r')
        except (IOError, OSError):
            # it is ok for file to not exist on system
            retval = None
        except PermissionError:
            self._console.exception(constants.ERR_FILE_PERM+default_path)
        return retval

    def _get_user_cred_file(self, in_cred_path):
        """Get a handle to the credential file, if one exists, which the
           user specified as an option.
           Returns:
           File handle.
           Throws:
           Exception if file does not exist.
        """
        retval = None
        try:
            retval = open(in_cred_path, 'r')
        except (IOError, OSError):
            # file must exist and be readable
            self._console.exception(constants.ERR_FILE_OPEN + in_cred_path)
        except PermissionError:
            self._console.exception(constants.ERR_FILE_PERM + in_cred_path)
        return retval

    def _read_cred_file(self, in_file):
        """Read the credentials from a credentials file.

        Function reads the organization, access key and user from a saved
        file and returns a dictionary with the values.
        Arguments:
        in_file -- File descriptor of credentials file.
        Returns:
        A dictionary object with the organization, access key and user.
        """
        assert in_file is not None
        indata = None
        try:
            if in_file is not None:
                with in_file as infile:
                    indata = json.load(infile)
                    jsonschema.validate(indata, constants.CREDENTIALS_SCHEMA)
            else:
                self._console.exception('Internal: None passed to method.')
        except jsonschema.exceptions.ValidationError as e:
            self._console.exception(constants.ERR_BAD_CRED_FILE)
        if indata['version'] != constants.MAX_VERSION_CREDENTIALS:
            self._console.exception(constants.ERR_CRED_FILE_VERSION)
        assert indata is not None
        return indata



# Copyright (c) 2018 UniquID

import json
import uniquid.core.constants as constants

# constants
_TAB = '\t'


class CliConsole:
    """Generic object which controls all interactions with the command line and
    the console. Independant of the framework which the tool is built with.
    """

    def __init__(self,
                 in_ok=None,
                 in_error=None,
                 in_output=constants.FORMAT_TEXT,
                 in_exc_ctor=RuntimeError):
        """Create a CliConsole object and store the references to the
           platform specific functions.

           Arguments:
           in_ok -- Framework function to print strings to console.
           in_error -- Framework function to print errors to the console.
           in_output -- Format of output to terminal.
           in_exc_ctor -- Reference to contructor function to create
           exceptions which should be thrown for the framework."""
        self._print_ok = None
        self._print_error = None
        self._raise_exception = None
        self._format = in_output
        self.is_list_mode = False
        self.num_prints = 0
        if in_ok is not None:
            self._print_ok = in_ok
        elif in_ok is not None:
            self._print_warning = in_ok
        if in_error is not None:
            self._print_error = in_error
        elif in_ok is not None:
            self._print_error = in_ok
        self._excp_ctor = in_exc_ctor

    def ok(self, in_string):
        """Print a string using the pre-configured print function."""
        if self._print_ok is None:
            self.exception('Print function not assigned.')
        if (self._format == constants.FORMAT_TEXT):
            self._print_ok(in_string)
        elif (self._format == constants.FORMAT_JSON):
            if (self.is_list_mode and
                    self.num_prints > 0):
                self._print_ok(',')
            self._print_ok('{"message": "' + in_string + '"}')
        else:
            self.exception('Internal error: Unexpected format type')
        self.num_prints = self.num_prints + 1

    def error(self, in_string):
        """Print a string using the pre-configured print function."""
        if self._print_error is None:
            self.exception('Print error function not assigned.')
        if (self._format == constants.FORMAT_TEXT):
            self._print_error('Error: ' + in_string)
        elif (self._format == constants.FORMAT_JSON):
            if (self.is_list_mode and
                    self.num_prints > 0):
                self._print_ok(',')
            self._print_error('{"error": "' + in_string + '"}')
        else:
            self.exception('Internal error: Unexpected format type')
        self.num_prints = self.num_prints + 1

    def exception(self, in_string):
        """Throw an exception which contains the message."""
        raise self._excp_ctor(in_string)

    def begin_list(self):
        """Print prefix lines to all text output for the command.

        Places the object in List mode, which means it expects to print
        multiple data objects to the CLI."""
        if self._print_ok is None:
            self.exception('Print function not assigned.')
        if self._format == constants.FORMAT_JSON:
            self._print_ok('[')
        self.is_list_mode = True

    def end_list(self):
        """Print suffix lines to all text output for the command."""
        if self._print_ok is None:
            self.exception('Print function not assigned.')
        if self._format == constants.FORMAT_JSON:
            self._print_ok(']')
        self.is_list_mode = False
        self.num_prints = 0

    def print_objects(self, in_objs):
        """Print object(s) to the console in the requested format.

        Arguments:
        in_objs -- A Python object. May be a dictionary or a list.
        """
        if self._format == constants.FORMAT_JSON:
            if (self.is_list_mode and
                    self.num_prints > 0):
                self._print_ok(',')
            self._print_ok(json.dumps(in_objs,
                                      sort_keys=False,
                                      indent=4))
        elif self._format == constants.FORMAT_TEXT:
            if isinstance(in_objs, list):
                # print the header from the first object in the list
                if (len(in_objs) > 0 and
                        isinstance(in_objs[0], dict)):
                    key_list = self._get_hier_keys(in_objs[0])
                    self._print_ok(_TAB.join(key_list).upper())
                # print rows for all objects in the list
                for obj in in_objs:
                    self._print_ok(_TAB.join(self._get_hier_values(obj)))
            elif isinstance(in_objs, dict):
                # print the header from the object
                self._print_ok(_TAB.join(self._get_hier_keys(in_objs)).upper())
                # print row with values from object
                self._print_ok(_TAB.join(self._get_hier_values(in_objs)))
            else:
                self.exception('Internal: Unsupported object type: '
                               + str(type(in_objs)))
        else:
            self.exception('Internal: Unsupported text format.')
        self.num_prints = self.num_prints + 1

    def _get_hier_keys(self, in_objs):
        """Get a list of all of the keys in the object hierarchy."""
        retval = list()
        if isinstance(in_objs, list):
            for item in in_objs:
                list_keys = self._get_hier_keys(item)
                retval.extend(list_keys)
        elif isinstance(in_objs, dict):
            for key, value in in_objs.items():
                if (isinstance(value, dict) or
                        isinstance(value, list)):
                    sub_keys = self._get_hier_keys(value)
                    if len(sub_keys) > 0:
                        for sub_key in sub_keys:
                            retval.append(str(key) + '_' + str(sub_key))
                    else:
                        retval.append(str(key))
                else:
                    retval.append(str(key))
        return retval

    def _get_hier_values(self, in_objs):
        """Get a list of all of the values in the object hierarchy."""
        retval = list()
        if isinstance(in_objs, list):
            aglom = list()
            aglom.append('[')
            for item in in_objs:
                list_values = self._get_hier_values(item)
                for value in list_values:
                    aglom.append(str(value))
                if item is not in_objs[-1]:
                    aglom.append(',')
            aglom.append(']')
            retval.append(''.join(aglom))
        elif isinstance(in_objs, dict):
            for item in in_objs.values():
                dict_values = self._get_hier_values(item)
                for item in dict_values:
                    retval.append(str(item))
        else:
            retval.append(str(in_objs))
        return retval



# Copyright (c) 2018 UniquID

"""Helper functions serving all modules.
"""

from operator import itemgetter


def parse_url(in_url, in_def_ip, in_def_port):
    """Parse a URL and extract the IP address and the port number.

    If the domain/IP address or port number is missing, use the default
    values.
    Arguments:
    in_url -- URL to be parsed.
    in_def_ip -- Default IP address to return. Must be a string.
    in_def_port -- Default port number to return. Must be a string.
    Returns:
    Returns a dictionary containing the IP address and port number.
    Throws:
    RuntimeError if an invalid default argument is passed to the function.
    """
    retval = {'ip': in_def_ip, 'port': in_def_port}
    tokens = []
    if (in_def_ip is None or
            in_def_port is None):
        raise RuntimeError('Internal. Null argument.')
    elif (not isinstance(in_def_ip, str) or
            not isinstance(in_def_port, str)):
        raise RuntimeError('Internal. Non-string argument.')
    if in_url:
        tokens = in_url.split(':')
    if len(tokens) > 1:
        # there was a colon in url
        if len(tokens[-1]) > 0:
            retval['port'] = tokens[-1]
        if len(tokens[-2]) > 0:
            retval['ip'] = tokens[-2].replace('/', '')
    elif (in_url and len(in_url) > 0):
        # no colon in url
        retval['ip'] = in_url.replace('/', '')
    return retval


def filter_list(in_list=None,
                in_key=None,
                in_value=None,
                in_is_case_sensitive=False,
                in_is_inclusive=False):
    """Filter a list of dictionaries based on the values of one field.

    A new list is returned and dictionaries are included in it if their field
    matches the specified value.  Inclusive matching means that dictionaries
    which are missing the field are also included in the returned list.
    Exclusive matching means that dictionaries which are missing the field are
    not included in the returned list.
    Arguments:
    in_list -- List of dictionary objects.
    in_key -- String name of the field.
    in_value -- Value the field should match.
    in_is_case_sensitive -- Boolean. True if filtering of values is case
    sensitive, otherwise False. Case insensitive by default.
    in_is_inclusive -- Boolean. True indicates that the matching should be
    inclusive, otherwise there is exclusive matching.
    Returns:
    A list of dictionary objects or an empty list.
    Throws:
    RuntimeError if an incorrect argument is passed.
    """
    retval = list()
    if not isinstance(in_list, list):
        raise RuntimeError('Internal. Incorrect argument type for in_list.')
    if (not in_key or
            not isinstance(in_key, str) or
            len(in_key) < 1):
        raise RuntimeError('Internal. Incorrect argument in_key.')
    if (not in_value or
            len(in_value) < 1):
        raise RuntimeError('Internal. Incorrect argument in_value.')
    for obj in in_list:
        if not isinstance(obj, dict):
            raise RuntimeError('Internal. List must contain dict objects.')
        if ((in_key in obj and
                ((in_is_case_sensitive and
                  obj[in_key] == in_value) or
                 ((not in_is_case_sensitive and
                   obj[in_key].lower() == in_value.lower())))) or
            (in_key not in obj and
                in_is_inclusive)):
            retval.append(obj)
    return retval


def sort_list(in_list=None,
              in_key=None,
              in_is_ascending=True,
              in_is_inclusive=True):
    """Sort a list of dictionaries based on the values of one field.

    A new list is returned and all of the dictionaries are included in it, in
    either ascending or descending order.  Inclusive matching means that
    dictionaries which are missing the field are also included in the
    returned list. Exclusive matching means that dictionaries which are missing
    the field are not included in the returned list.
    Arguments:
    in_list -- List of dictionary objects.
    in_key -- String name of the field.
    in_is_ascending -- True for ascending order, False for descending order.
    in_is_inclusive -- Boolean. True indicates that the sorting should be
    inclusive, otherwise dictionaries which are missing the key are
    excluded from the returned list.
    Returns:
    A list of dictionary objects or an empty list.
    Throws:
    RuntimeError if an incorrect argument is passed.
    """
    retval = list()
    missing_key = list()
    sort_list = list()
    if not isinstance(in_list, list):
        raise RuntimeError('Internal. Incorrect argument type for in_list.')
    if (not in_key or
            not isinstance(in_key, str) or
            len(in_key) < 1):
        raise RuntimeError('Internal. Incorrect argument in_key.')
    for item in in_list:
        if not isinstance(item, dict):
            raise RuntimeError('Internal. List must contain dict objects.')
        if in_key in item:
            sort_list.append(item)
        if (in_key not in item and
                in_is_inclusive):
            missing_key.append(item)
    if len(sort_list) > 0:
        sort_list = sorted(sort_list,
                           key=itemgetter(in_key),
                           reverse=not in_is_ascending)
    retval.extend(sort_list)
    retval.extend(missing_key)
    return retval



# Copyright (c) 2018 UniquID

import click
from uniquid.core.login_manager import LoginManager
from uniquid.core.cli_console import CliConsole
import uniquid.core.constants as constants
import uniquid.commands.common as common


@click.command()
@click.argument('org',
                required=False)
@click.option('--credentials-file',
              envvar='UNIQUID_CRED_FILE',
              help=('Path to a credentials file.'
                    ' Option is available to allow the user to specify the'
                    ' path to a file containing the user\'s login credentials.'
                    ' By default, if this option is not passed, the tool will'
                    ' search for the credential file at'
                    ' \".uniquid/credentials.json\" in the user\'s'
                    ' home directory.'
                    ' The path can also be set using the environment'
                    ' variable UNIQUID_CRED_FILE.'),
              required=False)
@click.option('--user',
              envvar='UNIQUID_USER',
              help=('UniquID Username. Provided during account registration.'
                    ' The Username must be provided if no crediantials file'
                    ' is provided.'
                    ' The option can also be set using the environment'
                    ' variable UNIQUID_USER.'),
              required=False)
@click.option('--access-key',
              envvar='UNIQUID_ACCESS_KEY',
              help=('UniquID Access Key. Provided after account registration.'
                    ' The Access Key must be provided if no crediantials file'
                    ' is provided.'
                    ' The option can also be set using the environment'
                    ' variable UNIQUID_ACCESS_KEY.'),
              required=False)
@click.option('--login-url',
              envvar='UNIQUID_LOGIN_URL',
              help=('UniquID server URL for the Login. Option is available to'
                    ' allow the user to specify the URL of the login server.'
                    ' This is not normally required. The URL'
                    ' must be in the format <ip address>:<port number>. The'
                    ' inclusion of the port number is optional.'
                    ' The option can also be set using the environment'
                    ' variable UNIQUID_LOGIN_URL.'),
              required=False)
@click.option('--org-url',
              envvar='UNIQUID_ORG_URL',
              help=('UniquID server URL for the user\'s Organization. Option'
                    ' is available to'
                    ' allow the user to specify the URL of their assigned'
                    ' server. This is not normally required. The URL'
                    ' must be in the format <ip address>:<port number>. The'
                    ' inclusion of the port number is optional.'
                    ' The option can also be set using the environment'
                    ' variable UNIQUID_ORG_URL.'),
              required=False)
def login(org, access_key, user, login_url, org_url, credentials_file):
    """Login to the Uniquid system, passing the organization identifier ORG.

    When first logging into the system, the user must pass the location of
    the credentials file they downloaded during registration, using the
    "--credentials-file" option. If they saved the credentials file at
    "<home directory>/.uniquid/credentials.json" then they do not need to
    pass the file path as the file will be automatically detected by the login
    command.
    If the user does not want to use the credentials file then they can
    pass the organization identifier, user and access key using command line
    options.  If the user specified the organization on the command line, then
    after the first successful login, they no longer need to pass the
    organization unless they want to change the organization account which
    they are using.
    If no credentials file is being used, the user must always pass the
    user identifier and the access key.
    """
    cc = CliConsole(click.echo, common.print_error,
                    constants.FORMAT_TEXT, click.ClickException)
    lm = LoginManager(cc)
    lm.connect(org, access_key, user, login_url, org_url, credentials_file)


@click.command()
def logout():
    """Logout from the Uniquid system."""
    cc = CliConsole(click.echo, common.print_error,
                    constants.FORMAT_TEXT, click.ClickException)
    lm = LoginManager(cc)
    lm.disconnect()


@click.command()
@click.option('--verbose',
              is_flag=True,
              default=False,
              help=('Print more detailed status information about'
                    ' the connection to the Uniquid system.'),
              required=False)
@click.option('--output',
              default=constants.FORMAT_TEXT,
              type=click.Choice(constants.FORMAT_ALL),
              help=('Format used to print data to the console. Valid options'
                    ' are: ') + str(constants.FORMAT_ALL),
              required=False)
def status(verbose, output):
    """Print the status of the connection to Uniquid system."""
    cc = CliConsole(click.echo, common.print_error,
                    output, click.ClickException)
    lm = LoginManager(cc)
    state = lm.get_conn_state()
    cc.begin_list()
    cc.ok(constants.TXT_CONN_STATUS + state)
    if (verbose and
            state == constants.TXT_CONN_LOGGED_IN):
        lm.print_info()
    cc.end_list()



# Copyright (c) 2018 UniquID

import json
import requests
import jsonschema
import uniquid.core.constants as constants
import uniquid.core.schema as schema


class Contracts:
    """Implements all commands related to contracts.
    """

    def __init__(self, in_console, in_login_mgr):
        """Initialise an instance of the class."""
        self._console = in_console
        self._login_mgr = in_login_mgr

    def create_contracts(self, in_json):
        """Create new contracts.

            Arguments:
            in_json -- JSON sting with a list of contract objects.
        """
        if in_json is None:
            self._console.exception(constants.exception('Internal. ' +
                                                        'Missing JSON.'))
        # check that user is logged in
        session_key = self._login_mgr.get_session_key()
        if session_key is None:
            self._console.exception(constants.ERR_LOGGED_OUT)
        # validate json before bothering the backend
        json_data = None
        try:
            json_data = json.loads(in_json)
            jsonschema.validate(json_data, schema.NEW_CONTRACT_SCHEMA)
        except (jsonschema.exceptions.ValidationError,
                json.decoder.JSONDecodeError) as e:
            self._console.exception(constants.ERR_INVALID_JSON + ': ' + str(e))
        # post json to server
        url_string = 'http://{0}:{1}/api/v1/contracts'
        url = url_string.format(self._login_mgr.get_api_ip(),
                                self._login_mgr.get_api_port())
        cookies = dict(JSESSIONID=session_key)
        count = 0
        try:
            for obj in json_data:
                response = requests.post(url, cookies=cookies, json=obj)
                if response.status_code == 200:
                    count = count + 1
                elif response.status_code == 401:
                    self._console.warning(constants.TXT_CONTRACT_INTERRUPTED
                                          + str(count))
                    self._console.exception(constants.ERR_LOGIN_REJECTED)
                elif response.status_code == 500:
                    self._console.warning(constants.TXT_CONTRACT_INTERRUPTED
                                          + str(count))
                    self._console.exception(constants.ERR_SERVER_ERROR)
                else:
                    self._console.warning(constants.TXT_CONTRACT_INTERRUPTED
                                          + str(count))
                    self._console.exception(constants.ERR_UNKNOWN_ERROR + ' ['
                                            + str(response.status_code) + ']')
            self._console.ok(constants.TXT_CONTRACT_CREATED + str(count))
        except requests.exceptions.ConnectionError:
            self._console.warning(constants.TXT_CONTRACT_INTERRUPTED
                                  + str(count))
            self._console.exception(constants.ERR_NO_RESPONSE)
        self._console.ok(constants.TXT_CONTRACT_CREATED_ALL)

    def list_contracts(self):
        """List all contracts.
        """
        # check that user is logged in
        session_key = self._login_mgr.get_session_key()
        if session_key is None:
            self._console.exception(constants.ERR_LOGGED_OUT)
        # retrieve the device info from the server using HTTP API
        url_string = 'http://{0}:{1}/api/v1/contracts'
        url = url_string.format(self._login_mgr.get_api_ip(),
                                self._login_mgr.get_api_port())
        cookies = dict(JSESSIONID=session_key)
        try:
            response = requests.get(url, cookies=cookies)
            if (response.status_code == 200
                    and isinstance(response.json(), list)):
                self._console.print_objects(response.json())
            elif response.status_code == 401:
                self._console.exception(constants.ERR_LOGIN_REJECTED)
            elif response.status_code == 500:
                self._console.exception(constants.ERR_SERVER_ERROR)
            else:
                self._console.exception(constants.ERR_UNKNOWN_ERROR + ' ['
                                        + str(response.status_code) + ']')
        except requests.exceptions.ConnectionError:
            self._console.exception(constants.ERR_NO_RESPONSE)

    def show_contract(self, in_txid):
        """Show a single contract.

        Arguments:
        in_txid -- Transaction ID corresponding to the contract.
        """
        # check that user is logged in
        session_key = self._login_mgr.get_session_key()
        if session_key is None:
            self._console.exception(constants.ERR_LOGGED_OUT)
        # retrieve the device info from the server using HTTP API
        url_string = 'http://{0}:{1}/api/v1/contracts'
        url = url_string.format(self._login_mgr.get_api_ip(),
                                self._login_mgr.get_api_port())
        cookies = dict(JSESSIONID=session_key)
        try:
            response = requests.get(url, cookies=cookies)
            if (response.status_code == 200
                    and response.json()):
                all_list = response.json()
                matched = None
                for contract in all_list:
                    if in_txid == contract.get('txid'):
                        matched = contract
                        break
                if matched:
                    self._console.print_objects(matched)
            elif response.status_code == 401:
                self._console.exception(constants.ERR_LOGIN_REJECTED)
            elif response.status_code == 500:
                self._console.exception(constants.ERR_SERVER_ERROR)
            else:
                self._console.exception(constants.ERR_UNKNOWN_ERROR + ' ['
                                        + str(response.status_code) + ']')
        except requests.exceptions.ConnectionError:
            self._console.exception(constants.ERR_NO_RESPONSE)

    def delete_contract(self, in_json):
        """Create a new contract.

            Arguments:
            in_json -- JSON sting with array of TXIDS.
        """
        if in_json is None:
            self._console.exception(constants.exception('Internal. ' +
                                                        'Missing JSON.'))
        # check that user is logged in
        session_key = self._login_mgr.get_session_key()
        if session_key is None:
            self._console.exception(constants.ERR_LOGGED_OUT)
        # validate json before bothering the backend
        json_data = None
        try:
            json_data = json.loads(in_json)
            jsonschema.validate(json_data, schema.DELETE_CONTRACT_SCHEMA)
        except (jsonschema.exceptions.ValidationError,
                json.decoder.JSONDecodeError) as e:
            self._console.exception(constants.ERR_INVALID_JSON + ': ' + str(e))
        # post json to server, once per contract
        for txid in json_data:
            url_string = 'http://{0}:{1}/api/v1/contracts/{2}'
            url = url_string.format(self._login_mgr.get_api_ip(),
                                    self._login_mgr.get_api_port(),
                                    txid)
            cookies = dict(JSESSIONID=session_key)
            try:
                response = requests.delete(url,
                                           cookies=cookies,
                                           json=json_data)
                if response.status_code == 200:
                    self._console.ok(constants.TXT_CONTRACT_DELETED
                                     + str(txid))
                elif response.status_code == 401:
                    self._console.exception(constants.ERR_LOGIN_REJECTED)
                elif response.status_code == 404:
                    self._console.exception(constants.ERR_NO_CONTRACT)
                elif response.status_code == 500:
                    self._console.exception(constants.ERR_SERVER_ERROR)
                else:
                    self._console.exception(constants.ERR_UNKNOWN_ERROR + ' ['
                                            + str(response.status_code) + ']')
            except requests.exceptions.ConnectionError:
                self._console.exception(constants.ERR_NO_RESPONSE)
        self._console.ok(constants.TXT_CONTRACT_DELETED_ALL)



# Copyright (c) 2018 UniquID

import click

# short aliases for all commands
_aliases = dict()


def get_cmd_name(in_alias=None):
    """Get the full command name of an alias. If the argument is not
    the alias of a command or the full name then return None.

    Arguments:
    in_alias -- String with alias of a command or the full name of the
    command.

    Returns:
    String with the full command name or None if the alias does not map to
    a command.
    """
    retval = None
    if (in_alias and
            in_alias in _aliases.keys()):
        retval = _aliases.get(in_alias)
    elif (in_alias and
            in_alias in _aliases.values()):
        retval = in_alias
    return retval


def set_alias(in_alias=None, in_cmd_name=None):
    """Sets the alias for a full command name.

    Arguments:
    in_alias -- Alias name.
    in_cmd_name -- Full name of the command.

    Returns:
    True if the alias was set correctly, otherwise false.
    """
    retval = False
    if (in_alias and
            in_cmd_name and
            in_alias not in _aliases.keys()):
        _aliases[in_alias] = in_cmd_name
        retval = True
    return retval


@click.command(name='list-aliases')
def list_aliases():
    """List all of the aliases available for full command names."""
    click.echo('Available command aliases:')
    for key, value in _aliases.items():
        click.echo('    {0} {1}'.format(key, value))



# Copyright (c) 2018 UniquID

"""
Configuration file for setuptools and pip.
"""

import sys
import platform
from setuptools import setup, find_packages
from os import path
import uniquid.core.constants as constants

min_version = (str(constants.MIN_VERSION_MAJOR)
               + '.' + str(constants.MIN_VERSION_MINOR))

if (sys.version_info <
        (constants.MIN_VERSION_MAJOR, constants.MIN_VERSION_MINOR)):
    print('Current Python version: ' + platform.python_version())
    sys.exit('Incorrect version of Python installed. Python < ' +
             min_version + ' is not supported.')

current_dir = path.abspath(path.dirname(__file__))
with open(path.join(current_dir, 'USER.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='uniquid',
    version=constants.APP_VERSION,
    description='Uniquid command line administration tool.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Uniquid Inc.',
    author_email='hello@uniquid.com',
    maintainer='Michael McCarthy',
    maintainer_email='mmccarthy@uniquid.com',
    url='http://uniquid.com/',
    classifiers=['Development Status :: 3 - Alpha',
                 'Environment :: Console',
                 'Intended Audience :: System Administrators',
                 'License :: Other/Proprietary License',
                 'Natural Language :: English',
                 'Programming Language :: Python :: 3.6',
                 'Topic :: Security',
                 'Topic :: System :: Systems Administration'],
    packages=find_packages(),
    data_files=[('.', ['LICENSE', 'CHANGELOG.md', 'USER.md'])],
    include_package_data=True,
    python_requires='>=' + min_version,
    install_requires=[
        'click>=6.7',
        'requests>=2.19',
        'jsonschema>=2.6'
    ],
    entry_points='''
        [console_scripts]
        uniquid=uniquid.cli:cli_group
    ''',
)



# Copyright (c) 2018 UniquID

import click
import json
from uniquid.core.login_manager import LoginManager
from uniquid.core.cli_console import CliConsole
from uniquid.core.contracts import Contracts
import uniquid.core.constants as constants
import uniquid.commands.common as common


@click.command(name='create-contracts')
@click.option('--input-json',
              default=None,
              type=click.STRING,
              help=('JSON object describing the contract.'),
              required=False)
@click.option('--input-json-file',
              default=None,
              type=click.File(mode='r'),
              help=('Path to file containing a list of JSON'
                    ' contract objects.'),
              required=False)
def create_contracts(input_json, input_json_file):
    """Create a new contract."""
    if (input_json is None) == (input_json_file is None):
        raise click.ClickException(constants.ERR_SINGLE_OPTION)
    cc = CliConsole(click.echo, common.print_error,
                    constants.FORMAT_TEXT, click.ClickException)
    lm = LoginManager(cc)
    contracts = Contracts(cc, lm)
    json_string = input_json
    if input_json_file:
        json_string = input_json_file.read()
    contracts.create_contracts(json_string)


@click.command(name='list-contracts')
@click.option('--output',
              default=constants.FORMAT_TEXT,
              type=click.Choice(constants.FORMAT_ALL),
              help=('Format used to print data to the console. Valid options'
                    ' are: ') + str(constants.FORMAT_ALL),
              required=False)
def list_contracts(output):
    """Show a list of all the contracts."""
    cc = CliConsole(click.echo, common.print_error,
                    output, click.ClickException)
    lm = LoginManager(cc)
    contracts = Contracts(cc, lm)
    contracts.list_contracts()


@click.command(name='show-contract')
@click.argument('txid', nargs=1)
@click.option('--output',
              default=constants.FORMAT_TEXT,
              type=click.Choice(constants.FORMAT_ALL),
              help=('Format used to print data to the console. Valid options'
                    ' are: ') + str(constants.FORMAT_ALL),
              required=False)
def show_contract(txid, output):
    """Show a single contract, identified by transaction TXID."""
    cc = CliConsole(click.echo, common.print_error,
                    output, click.ClickException)
    lm = LoginManager(cc)
    contracts = Contracts(cc, lm)
    contracts.show_contract(txid)


@click.command(name='delete-contracts')
@click.argument('txids', nargs=-1)  # variadic
@click.option('--input-json',
              default=None,
              type=click.STRING,
              help=('JSON array with TXIDs of contracts to be deleted.'),
              required=False)
@click.option('--input-json-file',
              default=None,
              type=click.File(mode='r'),
              help=('Path to file containing a JSON array of the TXIDs'
                    ' of the contracts to be deleted.'),
              required=False)
def delete_contracts(txids, input_json, input_json_file):
    """Delete all of the contracts which are identified by their
        transaction IDs being in the argument list TXIDS."""
    if ((input_json is None) == (input_json_file is None) and
            (txids is None or len(txids) == 0)):
        raise click.ClickException(constants.ERR_SINGLE_OPTION)
    cc = CliConsole(click.echo, common.print_error,
                    constants.FORMAT_TEXT, click.ClickException)
    lm = LoginManager(cc)
    contracts = Contracts(cc, lm)
    json_string = input_json
    if input_json_file:
        json_string = input_json_file.read()
    # if user passed one+ arguments, these take priority over a JSON array.
    if(len(txids) > 0):
        txid_list = list()
        for txid in txids:
            txid_list.append(str(txid))
        json_string = json.dumps(txid_list)
    contracts.delete_contract(json_string)



# Copyright (c) 2018 UniquID

import click

from uniquid.commands.login import login
from uniquid.commands.login import logout
from uniquid.commands.login import status
from uniquid.commands.device import show_device
from uniquid.commands.device import list_devices
from uniquid.commands.contract import create_contracts
from uniquid.commands.contract import list_contracts
from uniquid.commands.contract import show_contract
from uniquid.commands.contract import delete_contracts
from uniquid.commands.share import list_shares
from uniquid.commands.share import create_shares
from uniquid.commands.share import delete_shares
from uniquid.commands.aliases import list_aliases
import uniquid.core.constants as constants
import uniquid.commands.aliases as aliases


class AliasGroup(click.Group):
    """Sub-class of Group allows aliases to be defined for each of the
    commands.
    """
    def get_command(self, ctx, cmd_name):
        full_name = aliases.get_cmd_name(cmd_name)
        if full_name is None:
            full_name = cmd_name
        return click.Group.get_command(self, ctx, full_name)


# high-level grouping of all commands
@click.command(cls=AliasGroup)
@click.version_option(constants.APP_VERSION)
def cli_group():
    pass


# add each command to the high-level group
cli_group.add_command(list_aliases)
cli_group.add_command(login)
cli_group.add_command(logout)
cli_group.add_command(status)
cli_group.add_command(show_device)
cli_group.add_command(list_devices)
cli_group.add_command(create_contracts)
cli_group.add_command(list_contracts)
cli_group.add_command(show_contract)
cli_group.add_command(delete_contracts)
cli_group.add_command(list_shares)
cli_group.add_command(create_shares)
cli_group.add_command(delete_shares)

# configure aliases for the commands which have an alias
aliases.set_alias('li', login.name)
aliases.set_alias('lo', logout.name)
aliases.set_alias('st', status.name)
aliases.set_alias('la', list_aliases.name)
aliases.set_alias('sd', show_device.name)
aliases.set_alias('ld', list_devices.name)
aliases.set_alias('cc', create_contracts.name)
aliases.set_alias('lc', list_contracts.name)
aliases.set_alias('sc', show_contract.name)
aliases.set_alias('dc', delete_contracts.name)
aliases.set_alias('ls', list_shares.name)
aliases.set_alias('cs', create_shares.name)
aliases.set_alias('ds', delete_shares.name)


if __name__ == '__main__':
    cli_group()
